
// Additions for easier use with B3D

#ifdef _MSC_VER
#pragma warning(disable:4291)  // for VC++, no complaints about "no matching operator delete found"
#endif

#include "objects.h"
#include <ode/ode.h>
#include "joints/joints.h"
#include <ode/odemath.h>
#include <ode/matrix.h>
#include "step.h"
#include "quickstep.h"
#include "util.h"
#include <ode/memory.h>
#include <ode/error.h>
#include "collision_kernel.h"

// misc defines
#define ALLOCA dALLOCA16
#define RADTODEG (float)57.29578

/*
 * This wrapper is intended to provide a massively simplified version of ODE that can
 * simply plug straight into Blitz3D (or any other engine) applications to provide very
 * simple physical effects like falling boxes (yay!) and ragdolls. It is not intended
 * for any kind of intensive or accurate simulation and does not provide access to
 * many of ODE's key features, so don't use it if you need a "serious" physics solution.
 */

/****************************************************************************************
Static objects
****************************************************************************************/

dWorldID miniWorld;
int miniUpdateMethod;
int miniMaxContacts;
#define mini_MAXCONTACTS 256	//If this isn't enough, consider not doing whatever you're doing
dSpaceID miniSpace;
dJointGroupID miniContacts;

struct miniMaterial {
	int mode;
	float mu, bounce, bounceVel;
	float softERP, softCFM;
};

struct miniParams {
	float ERP, CFM;
	int doAutoDisable;
	float adLinearThreshold, adAngularThreshold;
	int adSteps;
	float adTime;
	int quickStepNI;
	float quickStepW;
	float linearDampingScale, angularDampingScale;
	float linearDampingThreshold, angularDampingThreshold;
	float maxAngSpeed;
	float cMaxCorrectingVel;
	float cSurfaceLayer;
	miniMaterial * defaultMaterial;
	int updateMethod, maxContacts;
};

struct miniAutoD {
	int flag;
	float linearThreshold, angularThreshold;
	int steps;
	float time;
};

struct miniDamp {
	float linearScale, angularScale;
	float linearThreshold, angularThreshold;
	float maxAngularSpeed;
};

struct miniTriMesh {
	dGeomID geom;
	struct miniTriMesh * next;
	struct miniTriMesh * prev;
};

struct miniTriMesh * mini_TriMeshFirst;
struct miniTriMesh * mini_TriMeshLast;


/****************************************************************************************
Wrapper functions
****************************************************************************************/

#define BBCALL _stdcall
extern "C" {

/****************************************************************************************
World functions
****************************************************************************************/

// Create a new simulation
ODE_API void BBCALL ODE_CreateSimulation_()
{
	dInitODE();
	miniWorld = dWorldCreate();
	miniSpace = dHashSpaceCreate(0);
	miniContacts = dJointGroupCreate(0);
	mini_TriMeshFirst = NULL;
	mini_TriMeshLast = NULL;
}

// Free a simulation
ODE_API void BBCALL ODE_FreeSimulation_()
{
	dSpaceDestroy(miniSpace);
	dJointGroupEmpty(miniContacts);
	dJointGroupDestroy(miniContacts);
	dWorldDestroy(miniWorld);
	dCloseODE();
}

// Change the gravity
ODE_API void BBCALL ODE_SetGravity(float gx, float gy, float gz)
{
	dWorldSetGravity(miniWorld, gx, gy, gz);
}

// Set the various other global options
ODE_API void BBCALL ODE_SetGlobalParameters(miniParams * mP)
{
	dWorldSetERP(miniWorld, mP->ERP);
	dWorldSetCFM(miniWorld, mP->CFM);
	dWorldSetAutoDisableFlag(miniWorld, mP->doAutoDisable);
	dWorldSetAutoDisableLinearThreshold(miniWorld, mP->adLinearThreshold);
	dWorldSetAutoDisableAngularThreshold(miniWorld, mP->adAngularThreshold);
	dWorldSetAutoDisableSteps(miniWorld, mP->adSteps);
	dWorldSetAutoDisableTime(miniWorld, mP->adTime);
	dWorldSetQuickStepNumIterations(miniWorld, mP->quickStepNI);
	dWorldSetQuickStepW(miniWorld, mP->quickStepW);
	dWorldSetLinearDamping(miniWorld, mP->linearDampingScale);
	dWorldSetAngularDamping(miniWorld, mP->angularDampingScale);
	dWorldSetMaxAngularSpeed(miniWorld, mP->maxAngSpeed);
	dWorldSetContactMaxCorrectingVel(miniWorld, mP->cMaxCorrectingVel);
	dWorldSetContactSurfaceLayer(miniWorld, mP->cSurfaceLayer);

	miniUpdateMethod = mP->updateMethod;
	miniMaxContacts = mP->maxContacts;
}

// Get the global options and put them in a Blitz object
ODE_API void BBCALL ODE_GetGlobalParameters(miniParams * mP)
{
	mP->ERP = dWorldGetERP(miniWorld);
	mP->CFM = dWorldGetCFM(miniWorld);
	mP->doAutoDisable = dWorldGetAutoDisableFlag(miniWorld);
	mP->adLinearThreshold = dWorldGetAutoDisableLinearThreshold(miniWorld);
	mP->adAngularThreshold = dWorldGetAutoDisableAngularThreshold(miniWorld);
	mP->adSteps = dWorldGetAutoDisableSteps(miniWorld);
	mP->adTime = dWorldGetAutoDisableTime(miniWorld);
	mP->quickStepNI = dWorldGetQuickStepNumIterations(miniWorld);
	mP->quickStepW = dWorldGetQuickStepW(miniWorld);
	mP->linearDampingScale = dWorldGetLinearDamping(miniWorld);
	mP->angularDampingScale = dWorldGetAngularDamping(miniWorld);
	mP->maxAngSpeed = dWorldGetMaxAngularSpeed(miniWorld);
	mP->cMaxCorrectingVel = dWorldGetContactMaxCorrectingVel(miniWorld);
	mP->cSurfaceLayer = dWorldGetContactSurfaceLayer(miniWorld);

	mP->updateMethod = miniUpdateMethod;
	mP->maxContacts = miniMaxContacts;
}

// Private function for doing the magical stuff
void nearCallBack(void *data, dGeomID o1, dGeomID o2)
{
	dBodyID b1 = dGeomGetBody(o1);
    dBodyID b2 = dGeomGetBody(o2);
	if (b1 == b2) return;

	dContact contact[mini_MAXCONTACTS];
	miniMaterial * m1 = (miniMaterial *)dGeomGetData(o1);
	miniMaterial * m2 = (miniMaterial *)dGeomGetData(o2);

	int i;
	for ( i = 0 ; i < miniMaxContacts ; i++ ) {
		contact[i].surface.mode = m1->mode | m2->mode;
		contact[i].surface.mu = (m1->mu + m2->mu) / 2;
		contact[i].surface.bounce = (m1->bounce + m2->bounce) / 2;
		contact[i].surface.bounce_vel = (m1->bounceVel + m2->bounceVel) / 2;
		contact[i].surface.soft_cfm = (m1->softCFM + m2->softCFM) / 2;
		contact[i].surface.soft_erp = (m1->softERP + m2->softERP) / 2;
	}

	if (int numc = dCollide (o1, o2, miniMaxContacts, &contact[0].geom, sizeof(dContact))) {
		for ( i = 0 ; i < numc ; i++ ) {
			dJointID c = dJointCreateContact (miniWorld, miniContacts, contact + i);
			dJointAttach (c, b1, b2);
		}
    }
}

// Taken from the examples
void setCurrentTransform(dGeomID geom)
{
	const dReal* Pos = dGeomGetPosition(geom);
	const dReal* Rot = dGeomGetRotation(geom);

	const dReal Transform[16] = 
	{
		Rot[0], Rot[4], Rot[8],  0,
		Rot[1], Rot[5], Rot[9],  0,
		Rot[2], Rot[6], Rot[10], 0,
		Pos[0], Pos[1], Pos[2],  1
	};
	dGeomTriMeshSetLastTransform( geom, *(dMatrix4*)(&Transform) );
}

// Step the simulation, using whichever method was set
ODE_API void BBCALL ODE_UpdatePhysics(float stepsize)
{
	//struct miniTriMesh * triMesh = mini_TriMeshFirst;
	dSpaceCollide(miniSpace, 0, &nearCallBack);
	//while (triMesh != NULL) {
	//	setCurrentTransform(triMesh->geom);
	//	triMesh = triMesh->next;
	//}
	if (miniUpdateMethod == 1) {
		dWorldQuickStep(miniWorld, stepsize);
	} else {
		dWorldStep(miniWorld, stepsize);
	}
	dJointGroupEmpty(miniContacts);
}


/****************************************************************************************
Rigid Body functions
****************************************************************************************/

//Creation and attributes
//=======================

ODE_API dBodyID BBCALL ODE_CreateBody_(miniMaterial * data)
{
	dBodyID newB = dBodyCreate(miniWorld);
	dBodySetData(newB, data);
	return newB;
}

//ODE_API void BBCALL ODE_FreeBody_(dBodyID body)
//{
//	dBodyDestroy(body);
//}

ODE_API void BBCALL ODE_SetMode(dBodyID body, int mode)
{
	switch (mode) {
		case 1:
			if (dBodyIsKinematic(body)) dBodySetDynamic(body);
			if (!dBodyIsEnabled(body)) dBodyEnable(body);
			break;
		case 2:
			if (!dBodyIsEnabled(body)) dBodyEnable(body);
			if (!dBodyIsKinematic(body)) dBodySetKinematic(body);
			break;
		case 3:
			if (dBodyIsEnabled(body)) dBodyDisable(body);
			break;
	}
}

ODE_API int BBCALL ODE_GetMode(dBodyID body)
{
	if (dBodyIsKinematic(body)) {
		return 2;
	} else if (dBodyIsEnabled(body)) {
		return 1;
	} else {
		return 3;
	}
}

ODE_API void BBCALL ODE_SetAutoDisable(dBodyID body, miniAutoD * AD)
{
	dBodySetAutoDisableFlag(body, AD->flag);
	dBodySetAutoDisableLinearThreshold(body, AD->linearThreshold);
	dBodySetAutoDisableAngularThreshold(body, AD->angularThreshold);
	dBodySetAutoDisableSteps(body, AD->steps);
	dBodySetAutoDisableTime(body, AD->time);
}

ODE_API void BBCALL ODE_GetAutoDisable(dBodyID body, miniAutoD * AD)
{
	AD->flag = dBodyGetAutoDisableFlag(body);
	AD->linearThreshold = dBodyGetAutoDisableLinearThreshold(body);
	AD->angularThreshold = dBodyGetAutoDisableAngularThreshold(body);
	AD->steps = dBodyGetAutoDisableSteps(body);
	AD->time = dBodyGetAutoDisableTime(body);
}

ODE_API void BBCALL ODE_SetDamping(dBodyID body, miniDamp * DS)
{
	dBodySetLinearDamping(body, DS->linearScale);
	dBodySetAngularDamping(body, DS->angularScale);
	dBodySetLinearDampingThreshold(body, DS->linearThreshold);
	dBodySetAngularDampingThreshold(body, DS->angularThreshold);
	dBodySetMaxAngularSpeed(body, DS->maxAngularSpeed);
}

ODE_API void ODE_GetDamping(dBodyID body, miniDamp * DS)
{
	DS->linearScale = dBodyGetLinearDamping(body);
	DS->angularScale = dBodyGetAngularDamping(body);
	DS->linearThreshold = dBodyGetLinearDampingThreshold(body);
	DS->angularThreshold = dBodyGetAngularDampingThreshold(body);
	DS->maxAngularSpeed = dBodyGetMaxAngularSpeed(body);
}

//ODE_API void BBCALL ODE_SetGravMode(dBodyID body, int mode)
//{
//	dBodySetGravityMode(body, mode);
//}

//ODE_API int BBCALL ODE_GetGravMode(dBodyID body)
//{
//	return dBodyGetGravityMode(body);
//}

ODE_API void BBCALL ODE_SetMass(dBodyID body, float mass)
{
//	dMass mS;
//	dMassSetZero(&mS);
//	mS.mass = mass;
//	dBodySetMass(body, &mS);
	body->mass.mass = mass;
}

ODE_API float BBCALL ODE_GetMass(dBodyID body)
{
	return body->mass.mass;
}


// Forces and movement
//=====================

ODE_API void BBCALL ODE_SetPosition(dBodyID b, float x, float y, float z)
{
	dBodySetPosition(b, x, y, z);
}

ODE_API void BBCALL ODE_SetRotation(dBodyID b, float pitch, float yaw, float roll)
{
	dMatrix3 R;
	dRFromEulerAngles(R, -pitch / RADTODEG, yaw / RADTODEG, -roll / RADTODEG);
	dBodySetRotation(b, R);
}

ODE_API float BBCALL ODE_GetX(dBodyID b)
{
	dAASSERT (b);
	return b->posr.pos[0];
}

ODE_API float BBCALL ODE_GetY(dBodyID b)
{
	dAASSERT (b);
	return b->posr.pos[1];
}

ODE_API float BBCALL ODE_GetZ(dBodyID b)
{
	dAASSERT (b);
	return b->posr.pos[2];
}

ODE_API void BBCALL ODE_GetRotation_(void * obj, int objType, float * matrixBank)
{
	const dReal * R;
	switch (objType) {
		case 1:
			R = dBodyGetRotation((dBodyID)obj);
			break;
		case 2:
			R = dGeomGetOffsetRotation((dGeomID)obj);
			break;
	}
	matrixBank[0] = R[0];
	matrixBank[1] = R[4];
	matrixBank[2] = R[8];
	matrixBank[3] = R[1];
	matrixBank[4] = R[5];
	matrixBank[5] = R[9];
}

ODE_API float BBCALL ODE_GetLinearVelX(dBodyID b)
{
	dAASSERT (b);
	return b->lvel[0];
}

ODE_API float BBCALL ODE_GetLinearVelY(dBodyID b)
{
	dAASSERT (b);
	return b->lvel[1];
}

ODE_API float BBCALL ODE_GetLinearVelZ(dBodyID b)
{
	dAASSERT (b);
	return b->lvel[2];
}

ODE_API float BBCALL ODE_GetAngularVelX(dBodyID b)
{
	dAASSERT (b);
	return b->avel[0];
}

ODE_API float BBCALL ODE_GetAngularVelY(dBodyID b)
{
	dAASSERT (b);
	return b->avel[1];
}

ODE_API float BBCALL ODE_GetAngularVelZ(dBodyID b)
{
	dAASSERT (b);
	return b->avel[2];
}

ODE_API float BBCALL ODE_GetForceX(dBodyID b)
{
	dAASSERT (b);
	return b->facc[0];
}

ODE_API float BBCALL ODE_GetForceY(dBodyID b)
{
	dAASSERT (b);
	return b->facc[1];
}

ODE_API float BBCALL ODE_GetForceZ(dBodyID b)
{
	dAASSERT (b);
	return b->facc[2];
}

ODE_API float BBCALL ODE_GetTorqueX(dBodyID b)
{
	dAASSERT (b);
	return b->tacc[0];
}

ODE_API float BBCALL ODE_GetTorqueY(dBodyID b)
{
	dAASSERT (b);
	return b->tacc[1];
}

ODE_API float BBCALL ODE_GetTorqueZ(dBodyID b)
{
	dAASSERT (b);
	return b->tacc[2];
}

// Geometry
//==========

ODE_API dGeomID BBCALL ODE_AddSphere_(dBodyID b, float radius)
{
	dGeomID sphere = dCreateSphere(miniSpace, radius);
	dGeomSetBody(sphere, b);
	if (b != 0) dGeomSetData(sphere, dBodyGetData(b));
	return sphere;
}

ODE_API dGeomID BBCALL ODE_AddBox_(dBodyID b, float sx, float sy, float sz)
{
	dGeomID box = dCreateBox(miniSpace, sx, sy, sz);
	dGeomSetBody(box, b);
	if (b != 0) dGeomSetData(box, dBodyGetData(b));
	return box;
}

ODE_API dGeomID BBCALL ODE_AddCylinder_(dBodyID b, float radius, float length)
{
	dGeomID cylinder = dCreateCylinder(miniSpace, radius, length);
	dGeomSetBody(cylinder, b);
	if (b != 0) dGeomSetData(cylinder, dBodyGetData(b));
	return cylinder;
}

ODE_API dGeomID BBCALL ODE_AddCapsule_(dBodyID b, float radius, float length)
{
	dGeomID capsule = dCreateCapsule(miniSpace, radius, length);
	dGeomSetBody(capsule, b);
	if (b != 0) dGeomSetData(capsule, dBodyGetData(b));
	return capsule;
}

void mini_addTriMesh(dGeomID triMesh)
{
	struct miniTriMesh * newTriMesh = new miniTriMesh;
	newTriMesh->geom = triMesh;
	
	newTriMesh->next = NULL;
	if (mini_TriMeshFirst == NULL) {
		mini_TriMeshFirst = newTriMesh;
		newTriMesh->prev = NULL;
	} else {
		newTriMesh->prev = mini_TriMeshLast;
		mini_TriMeshLast->next = newTriMesh;
	}
	mini_TriMeshLast = newTriMesh;
}

ODE_API dGeomID BBCALL ODE_AddTrimesh_(dBodyID b, const void * vertices, int vcount, const void * indices, int icount, const void * normals)
{
	dGeomID triMesh;
	dTriMeshDataID tmData;

	tmData = dGeomTriMeshDataCreate();
	dGeomTriMeshDataBuildSingle1(tmData, vertices, 3 * sizeof(float), vcount, indices, icount, 3 * sizeof(int), normals);
//	dGeomTriMeshDataBuildSingle(tmData, vertices, 3 * sizeof(float), vcount, indices, icount, 3 * sizeof(int));
	triMesh = dCreateTriMesh(miniSpace, tmData, NULL, NULL, NULL);
//	mini_addTriMesh(triMesh);
	dGeomSetBody(triMesh, b);
	if (b != 0) dGeomSetData(triMesh, dBodyGetData(b));

	return triMesh;
}

void mini_removeTriMesh(dGeomID triMeshGeom)
{
	struct miniTriMesh * triMesh = mini_TriMeshFirst;
	while (triMesh != NULL) {
		if (triMesh->geom == triMeshGeom) {
			if (triMesh->prev != NULL) triMesh->prev->next = triMesh->next;
			if (triMesh->next != NULL) triMesh->next->prev = triMesh->prev;
			if (triMesh == mini_TriMeshFirst) {
				if (triMesh->next == NULL) {
					mini_TriMeshFirst = NULL;
				} else {
					mini_TriMeshFirst = triMesh->next;
				}
			}
			if (triMesh == mini_TriMeshLast) {
				if (triMesh->prev == NULL) {
					mini_TriMeshLast = NULL;
				} else {
					mini_TriMeshLast = triMesh->prev;
				}
			}
			delete triMesh;
			return;
		}
		triMesh = triMesh->next;
	}
}

ODE_API dBodyID BBCALL ODE_FreeGeom_(dGeomID g)
{
	dBodyID b = dGeomGetBody(g);
	if (dGeomGetClass(g) == dTriMeshClass) {
//		mini_removeTriMesh(g);
		dTriMeshDataID triMeshData = dGeomTriMeshGetTriMeshDataID(g);
		dGeomTriMeshDataDestroy(triMeshData);
	}
	dGeomDestroy(g);
	return b;
}

ODE_API void BBCALL ODE_RotateGeom(dGeomID g, float pitch, float yaw, float roll)
{
	dMatrix3 R;
	dRFromEulerAngles(R, -pitch / RADTODEG, yaw / RADTODEG, -roll / RADTODEG);
	dGeomSetOffsetRotation(g, R);
}

ODE_API float BBCALL ODE_GeomX (dxGeom *g)
{
	dAASSERT (g);
	if (g->offset_posr)
	{
	  return g->offset_posr->pos[0];
	}
	return 0.0;
}

ODE_API float BBCALL ODE_GeomY (dxGeom *g)
{
	dAASSERT (g);
	if (g->offset_posr)
	{
	  return g->offset_posr->pos[1];
	}
	return 0.0;
}

ODE_API float BBCALL ODE_GeomZ (dxGeom *g)
{
	dAASSERT (g);
	if (g->offset_posr)
	{
	  return g->offset_posr->pos[2];
	}
	return 0.0;
}


/****************************************************************************************
Joint functions
****************************************************************************************/

ODE_API dJointID BBCALL ODE_CreateJoint_(int jointType, dBodyID bodyA, dBodyID bodyB)
{
	dJointID newJoint;
	switch (jointType) {
		case 1:	//Ball
			newJoint = dJointCreateBall(miniWorld, 0);
			break;
		case 2:	//Hinge
			newJoint = dJointCreateHinge(miniWorld, 0);
			break;
		case 3:	//Slider
			newJoint = dJointCreateSlider(miniWorld, 0);
			break;
	}
	dJointAttach(newJoint, bodyA, bodyB);
	return newJoint;
}

ODE_API int BBCALL ODE_Connected(dBodyID bodyA, dBodyID bodyB)
{
	return dAreConnectedExcluding(bodyA, bodyB, dJointTypeContact);
}

ODE_API void BBCALL ODE_PositionJoint(dJointID j, float px, float py, float pz)
{
	switch (dJointGetType(j)) {
		case 1:	//Ball
			dJointSetBallAnchor(j, px, py, pz);
			break;
		case 2:	//Hinge
			dJointSetHingeAnchor(j, px, py, pz);
			break;
	}
}

ODE_API float BBCALL ODE_JointX(dJointID j)
{
	float pVec[3];
	switch (dJointGetType(j)) {
		case 1:	//Ball
			dJointGetHingeAnchor(j, pVec);
			break;
		case 2:	//Hinge
			dJointGetBallAnchor(j, pVec);
			break;
	}
	return pVec[0];
}

ODE_API float BBCALL ODE_JointY(dJointID j)
{
	float pVec[3];
	switch (dJointGetType(j)) {
		case 1:	//Ball
			dJointGetHingeAnchor(j, pVec);
			break;
		case 2:	//Hinge
			dJointGetBallAnchor(j, pVec);
			break;
	}
	return pVec[1];
}

ODE_API float BBCALL ODE_JointZ(dJointID j)
{
	float pVec[3];
	switch (dJointGetType(j)) {
		case 1:	//Ball
			dJointGetHingeAnchor(j, pVec);
			break;
		case 2:	//Hinge
			dJointGetBallAnchor(j, pVec);
			break;
	}
	return pVec[2];
}

ODE_API void BBCALL ODE_SetJointAxis(dJointID j, float px, float py, float pz)
{
	if (dJointGetType(j) == dJointTypeHinge) {
		dJointSetHingeAxis(j, px, py, pz);
	}
}

ODE_API float BBCALL ODE_JointAxisX(dJointID j)
{
	float pVec[3];
	if (dJointGetType(j) == dJointTypeHinge) {
		dJointGetHingeAxis(j, pVec);
	}
	return pVec[0];
}

ODE_API float BBCALL ODE_JointAxisY(dJointID j)
{
	float pVec[3];
	if (dJointGetType(j) == dJointTypeHinge) {
		dJointGetHingeAxis(j, pVec);
	}
	return pVec[1];
}

ODE_API float BBCALL ODE_JointAxisZ(dJointID j)
{
	float pVec[3];
	if (dJointGetType(j) == dJointTypeHinge) {
		dJointGetHingeAxis(j, pVec);
	}
	return pVec[2];
}

ODE_API void BBCALL ODE_LimitJoint(dJointID j, float max, float min)
{
	switch (dJointGetType(j)) {
		case 2:	//Hinge
			dJointSetHingeParam(j, dParamHiStop, max);
			dJointSetHingeParam(j, dParamLoStop, min);
			break;
		case 3:	//Slider
			dJointSetSliderParam(j, dParamHiStop, max);
			dJointSetSliderParam(j, dParamLoStop, min);
			break;
	}
}

ODE_API float BBCALL ODE_JointMax(dJointID j)
{
	switch (dJointGetType(j)) {
		case 2:	//Hinge
			return dJointGetHingeParam(j, dParamHiStop);
			break;
		case 3:	//Slider
			return dJointGetSliderParam(j, dParamHiStop);
			break;
		default:
			return 0;
	}
}

ODE_API float BBCALL ODE_JointMin(dJointID j)
{
	switch (dJointGetType(j)) {
		case 2:	//Hinge
			return dJointGetHingeParam(j, dParamLoStop);
			break;
		case 3:	//Slider
			return dJointGetSliderParam(j, dParamLoStop);
			break;
		default:
			return 0;
	}
}


/***************************************************************************************/
}
