
;================================================;
;                                                ;
; miniODE example 2: World and object parameters ;
; ---------------------------------------------- ;
;                                                ;
;================================================;


;First things first: include miniODE's Blitz3D component in the project
Include "miniODE.bb"


;This needs to come before CreateSimulation
Graphics3D 1024,768,32,6

;Setup the Blitz3D world

Local SC_FPS=60		;Desired framerate - this is for "CPU breathing code"
Local VP_FPS=50		;Desired update rate - this mustn't change during the simulation
Local rtime=Floor(1000.0/SC_FPS),limited=True	;These variables all help with framerate control
Local phystime#=1000.0/VP_FPS
Local ctime=MilliSecs(),accum#=100,dtime

Local centre=CreatePivot()
Local camera = CreateCamera(centre)	;So we can see things happening!
PositionEntity camera,45,30,55
PointEntity camera,centre	;Point it at origin
Local light = CreateLight()
PositionEntity light,300,400,150
PointEntity light,centre


;Create a simulation and set the world's default gravity
ODE_CreateSimulation(0,-9.8,0)


;Let's create a floor again

;First the Blitz entity
Local gmesh = CreateCube()
ScaleMesh gmesh,40,1,40
EntityColor gmesh,150,150,150

;Then the ODE body...
Local gbase=ODE_CreateBody()	;Create it like you would a mesh
ODE_SetMode gbase,2			;This sets the body's mode to "kinematic" - it will be completely static and not affected by forces
Local gplane=ODE_AddBox(gbase,80,2,80)	;Add a box geom, like you would a surface to a mesh
ODE_SetEntity gbase,gmesh


;Let's adjust some parameters

;There are three kinds of parameters in miniODE: world settings, object damping, and object auto-disable.

;World settings control global issues - overall limits on damping and auto-disable, maximum object speeds,
;error correction and allowances, and overall simulation accuracy/speed tradeoffs.

;Damping refers to the amount ODE reduces the current velocity of an object by each step, which is not
;entirely unlike air resistance. It helps to maintain stability by stopping the system gaining energy
;when it shouldn't, but too much of it will make it look dead.

;Auto-disable is an optimisation feature that lets ODE ignore bodies that aren't contributing to the
;system right now. It also helps stability as if they aren't doing anything, they can't break anything!


;World settings are held in the ODE_WorldSettings type. By default one instance is created with these to
;initialise the world. To apply the settings in an ODE_WorldSettings instance, call ODE_SetGlobalParameters
;with it. You can call ODE_GetGlobalParameters on an instance to fill it with the current settings. This
;lets you store multiple world setups and switch between them quickly.

Local odeWorld.ODE_WorldSettings = New ODE_WorldSettings

ODE_GetGlobalParameters odeWorld	;This will take whatever the current settings are and store them in odeWorld

;Now we can change them if we want...
odeWorld\ERP=0.2				;This is the amount ODE will try to correct errors
odeWorld\CFM=0.0001				;This is a bit like softness - make it higher and the world turns to jelly
odeWorld\doAutoDisable=True		;Set to false if you don't want to use AutoDisable at all
odeWorld\adLinearThreshold=0.1	;These are global limits for autodisable
odeWorld\adAngularThreshold=0.1
odeWorld\adSteps=10
odeWorld\adTime=0
odeWorld\quickStepNI=100		;This is the number of iterations used by ODE_UpdatePhysics. A higher number gives
;								;a more accurate world, but is more computationally expensive.

odeWorld\quickStepW=1.3			;I would just leave this alone
odeWorld\linearDampingScale=0.1		;These are global limits for damping
odeWorld\angularDampingScale=0.1
odeWorld\linearDampingThreshold=0.1
odeWorld\angularDampingThreshold=0.1
odeWorld\maxAngSpeed=1000			;Maximum angular speed
odeWorld\cMaxCorrectingVel=1000		;Maximum velocity at which errors can be corrected
odeWorld\cSurfaceLayer=0.01		;How far bodies can penetrate each other. Allowing a little margin for error helps

odeWorld\defaultMaterial=ODE_DefaultMaterial	;The material (see example 3) that is applied to new bodies
odeWorld\updateMethod=1			;Whether to use WorldStep or WorldQuickStep - WorldStep is more accurate but slower
odeWorld\maxContacts=16;4		;Number of contact points to generate. More means more accuracy but at a speed cost

;And then upload them back into the simulation
ODE_SetGlobalParameters odeWorld

;Mainly you'll want to leave these alone. You can change the default world settings in ODE_CreateWorld and forget
;all about these options once you've found the settings you like.


;Object damping settings are held in the ODE_ObjectDampingSettings type. They are applied to an object upon
;creation and can also be accessed with ODE_SetDamping and ODE_GetDamping in the same way as the world settings.
;The default settings are held in the global ODE_DefaultDamping, which you can also just change in ODE_CreateWorld.

ODE_DefaultDamping = New ODE_ObjectDampingSettings
ODE_DefaultDamping\linearScale=0.02		;Changing these is more pointful on a per-object basis
ODE_DefaultDamping\angularScale=0.02
ODE_DefaultDamping\linearThreshold=0.1
ODE_DefaultDamping\angularThreshold=0.1
ODE_DefaultDamping\maxAngularVel=1000

;The same as above applies to auto-disable.
ODE_DefaultAutoDisable = New ODE_ObjectAutoDisableSettings
ODE_DefaultAutoDisable\flag=True
ODE_DefaultAutoDisable\linearThreshold=0.1
ODE_DefaultAutoDisable\angularThreshold=0.1
ODE_DefaultAutoDisable\steps=5
ODE_DefaultAutoDisable\time=0



;Now that's done, let's test the effects of the new settings by trying to stack some cubes
;With the default settings unchanged, they ought to stack quite happily and sit still
Local box1=ODE_CreateBody()
ODE_AddBox(box1,10,10,10)
ODE_SetMass box1,5
ODE_SetPosition box1,0,20,0
Local cube1=CreateCube()
ScaleEntity cube1,5,5,5
EntityColor cube1,255,0,0
ODE_SetEntity box1,cube1

Local box2=ODE_CreateBody()
ODE_AddBox(box2,10,10,10)
ODE_SetMass box2,5
ODE_SetPosition box2,0,40,0
Local cube2=CreateCube()
ScaleEntity cube2,5,5,5
EntityColor cube2,0,0,255
ODE_SetEntity box2,cube2



;The thing you;re most likely to want to do is set auto-disable on or off for individual objects.
ODE_GetAutoDisable box1,ODE_DefaultAutoDisable	;Get the current settings so we don't mess them up
ODE_DefaultAutoDisable\flag=True	;Change one or more
ODE_SetAutoDisable box1,ODE_DefaultAutoDisable	;Upload new settings to that specific object only


While Not KeyDown(1)
	dtime=(MilliSecs()-ctime)			;Time measurement (for frame rate control)
	accum=accum+dtime
	ctime=MilliSecs()
	
	While accum>=phystime	;This maintains constant interface speed regardless of video frame rate
		CaptureWorld		;Putting movement commands in this loop will mean they are executed at a constant rate
							;This is pretty much essential for a stable ODE simulation
		
		;Apply forces and manipulate the physics here, if necessary
		
		If KeyHit(57)		;Bounce the blue box
			ODE_SetMode box2,1	;Make sure it's awake
			ODE_ApplyForce box2,0,Rand(500,1000),0
		EndIf
		
		ODE_UpdatePhysics 0.05		;Actually update the physics simulation
		
		;If you're also using Blitz or ColDet collision detection you'll want to update the Blitz3D world here
		
		;UpdateWorld goes here, if you need it
		
		accum=accum-phystime
	Wend
	
	;If you're only using Blitz3D graphically, you can update it in the render part of the loop
	ODE_UpdateScene
	
	RenderWorld accum/phystime		;Render a tweened value to maintain smooth visual updates
	
	Text 10,10,"Press space to bounce the blue box"
	
	Delay rtime-(MilliSecs()-ctime)-(limited+1)		;Free any spare CPU time
	Flip limited
Wend

;Free everything and exit
ODE_FreeSimulation
End

;~IDEal Editor Parameters:
;~C#Blitz3D