
;=============================================
;                                            ;
; miniODE - a simple ODE wrapper for Blitz3D ;
;                                            ;
;=============================================


; Types
;=======

Type ODE_Material
	Field mode
	Field mu#,bounce#,bounceVel#
	Field softERP#,softCFM#
End Type

Type ODE_WorldSettings
	Field ERP#, CFM#
	Field doAutoDisable
	Field adLinearThreshold#, adAngularThreshold#
	Field adSteps
	Field adTime#
	Field quickStepNI
	Field quickStepW#
	Field linearDampingScale#, angularDampingScale#
	Field linearDampingThreshold#, angularDampingThreshold#
	Field maxAngSpeed#
	Field cMaxCorrectingVel#
	Field cSurfaceLayer#
	Field defaultMaterial.ODE_Material
	Field updateMethod, maxContacts
End Type

Type ODE_ObjectAutoDisableSettings
	Field flag
	Field linearThreshold#
	Field angularThreshold#
	Field steps
	Field time#
End Type

Type ODE_ObjectDampingSettings
	Field linearScale#
	Field angularScale#
	Field linearThreshold#
	Field angularThreshold#
	Field maxAngularVel#
End Type

Type ODE_Object
	Field ODEbody,B3Dentity
	Field geomBank
	Field jointBank
	Field material.ODE_Material
End Type

Type ODE_Geometry
	Field ODEgeom
	Field obj.ODE_Object
	Field material.ODE_Material
	Field vertexBank	;For trimeshes
	Field triangleBank
	Field normalBank
End Type


; Constants
;===========

Const ODE_CONTACT_BOUNCE = 4
Const ODE_CONTACT_SOFTERP = 8
Const ODE_CONTACT_SOFTCFM = 16

Const ODE_SPHERE_CLASS = 0
Const ODE_BOX_CLASS = 1
Const ODE_CAPSULE_CLASS = 2
Const ODE_CYLINDER_CLASS = 3
Const ODE_TRIMESH_CLASS = 8

Const ODE_JOINT_BALL = 1
Const ODE_JOINT_HINGE = 2
Const ODE_JOINT_SLIDER = 3


; Globals
;=========

Global ODE_DefaultMaterial.ODE_Material
Global ODE_DefaultDamping.ODE_ObjectDampingSettings
Global ODE_DefaultAutoDisable.ODE_ObjectAutoDisableSettings

Global private_ODE_MatrixBank, private_ODE_HelperPivot


; Functions
;===========

Function ODE_CreateSimulation(gx#, gy#, gz#)
	Local odeWorld.ODE_WorldSettings
	
	ODE_CreateSimulation_()
	ODE_SetGravity gx,gy,gz
	
	ODE_DefaultMaterial = ODE_CreateMaterial()
	
	ODE_DefaultAutoDisable = New ODE_ObjectAutoDisableSettings
	ODE_DefaultAutoDisable\flag=True
	ODE_DefaultAutoDisable\linearThreshold=0.1
	ODE_DefaultAutoDisable\angularThreshold=0.1
	ODE_DefaultAutoDisable\steps=5
	ODE_DefaultAutoDisable\time=0
	
	ODE_DefaultDamping = New ODE_ObjectDampingSettings
	ODE_DefaultDamping\linearScale=0.02
	ODE_DefaultDamping\angularScale=0.02
	ODE_DefaultDamping\linearThreshold=0.1
	ODE_DefaultDamping\angularThreshold=0.1
	ODE_DefaultDamping\maxAngularVel=1000
	
	odeWorld = New ODE_WorldSettings
	odeWorld\ERP=0.2
	odeWorld\CFM=0.0001
	odeWorld\doAutoDisable=True
	odeWorld\adLinearThreshold=0.1
	odeWorld\adAngularThreshold=0.1
	odeWorld\adSteps=10
	odeWorld\adTime=0
	odeWorld\quickStepNI=100
	odeWorld\quickStepW=1.3
	odeWorld\linearDampingScale=0.1
	odeWorld\angularDampingScale=0.1
	odeWorld\linearDampingThreshold=0.1
	odeWorld\angularDampingThreshold=0.1
	odeWorld\maxAngSpeed=1000
	odeWorld\cMaxCorrectingVel=1000
	odeWorld\cSurfaceLayer=0.01
	
	odeWorld\defaultMaterial=ODE_DefaultMaterial
	odeWorld\updateMethod=1
	odeWorld\maxContacts=16;4
	
	ODE_SetGlobalParameters odeWorld
	
	private_ODE_MatrixBank=CreateBank(24)
	private_ODE_HelperPivot=CreatePivot()
End Function

Function ODE_FreeSimulation()
	Local obj.ODE_Object,geom.ODE_Geometry,class
	
	Delete Each ODE_Material
	Delete Each ODE_ObjectAutoDisableSettings
	Delete Each ODE_ObjectDampingSettings
	Delete Each ODE_WorldSettings
	For obj=Each ODE_Object
		ODE_FreeObject(obj)
	Next
	For geom=Each ODE_Geometry
		ODE_FreeGeom geom\ODEgeom
	Next
	ODE_FreeSimulation_
	FreeBank private_ODE_MatrixBank
	FreeEntity private_ODE_HelperPivot
End Function

Function ODE_UpdateScene()
	Local obj.ODE_Object
	For obj=Each ODE_Object
		If obj\B3Dentity
			PositionEntity obj\B3Dentity,ODE_GetX(obj\ODEbody),ODE_GetY(obj\ODEbody),ODE_GetZ(obj\ODEbody)
			ODE_GetRotation obj\ODEbody,1,private_ODE_MatrixBank
			AlignToVector obj\B3Dentity,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
			AlignToVector obj\B3Dentity,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
		EndIf
	Next
End Function

Function ODE_CreateBody()
	Local newBody,obj.ODE_Object
	
	newBody=ODE_CreateBody_(ODE_DefaultMaterial)
	ODE_SetAutoDisable newBody,ODE_DefaultAutoDisable
	ODE_SetDamping newBody,ODE_DefaultDamping
	
	obj=New ODE_Object
	obj\ODEbody=newBody
	obj\geomBank=CreateBank()
	obj\jointBank=CreateBank()
	obj\material=ODE_DefaultMaterial
	
	Return newBody
End Function

Function ODE_FreeBody(body)
	Local obj.ODE_Object
	
	For obj=Each ODE_Object
		If obj\ODEbody=body Then Exit
	Next
	
	If obj=Null Then RuntimeError "Rigid body does not exist!"
	ODE_FreeObject(obj)
End Function

Function ODE_FreeObject(obj.ODE_Object)
	Local i,geom.ODE_Geometry
	
	For i=0 To BankSize(obj\geomBank)-4 Step 4
		geom=Object.ODE_Geometry(PeekInt(obj\geomBank,i))
		ODE_FreeGeom geom\ODEgeom
	Next
	
	For i=0 To BankSize(obj\jointBank)-4 Step 4
		ODE_FreeJoint(PeekInt(obj\jointBank,i))
	Next
	
	FreeBank obj\geomBank
	FreeBank obj\jointBank
	
	ODE_FreeBody_ obj\ODEbody
	Delete obj
End Function

Function ODE_CreateMaterial.ODE_Material(mode=ODE_CONTACT_BOUNCE,mu#=1.5,bounce#=0.1,bounceVel#=1,softERP#=0.2,softCFM#=0.0001)
	Local newMaterial.ODE_Material = New ODE_Material
	newMaterial\mode=mode
	newMaterial\mu=mu
	newMaterial\bounce=bounce
	newMaterial\bounceVel=bounceVel
	newMaterial\softERP=softERP
	newMaterial\softCFM=softCFM
	Return newMaterial
End Function

Function ODE_SetMaterial(ODEobj,material.ODE_Material)
	Local body.ODE_Object,geom.ODE_Geometry,i
	
	For geom=Each ODE_Geometry		;See if it's an individual geom
		If geom\ODEgeom=ODEobj
			geom\material=material
			ODE_SetGeomMaterial_ ODEobj,material
			Return
		EndIf
	Next
	
	For body=Each ODE_Object		;See if it's an entire object (ie. a rigid body)
		If body\ODEbody=ODEobj
			body\material=material
			ODE_SetBodyMaterial_ ODEobj,material
			For i=0 To BankSize(body\geomBank)-4	;Set the geom materials
				geom=Object.ODE_Geometry(PeekInt(body\geomBank,i))
				geom\material=material
				ODE_SetGeomMaterial_ geom\ODEgeom,material
			Next
			Return
		EndIf
	Next
End Function

Function ODE_SetEntity(ODEbody, B3Dentity)
	Local obj.ODE_Object
	For obj=Each ODE_Object
		If obj\ODEbody = ODEbody
			obj\B3Dentity = B3Dentity
			Return
		EndIf
	Next
End Function

Function ODE_GetPitch#(body)
	ODE_GetRotation body,1,private_ODE_MatrixBank
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
	Return EntityPitch(private_ODE_HelperPivot,1)
End Function

Function ODE_GetYaw#(body)
	ODE_GetRotation body,1,private_ODE_MatrixBank
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
	Return EntityYaw(private_ODE_HelperPivot,1)
End Function

Function ODE_GetRoll#(body)
	ODE_GetRotation body,1,private_ODE_MatrixBank
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
	Return EntityRoll(private_ODE_HelperPivot,1)
End Function

Function ODE_CountGeoms(body)
	Local obj.ODE_Object
	For obj=Each ODE_Object
		If obj\ODEbody=body Then Return BankSize(obj\geomBank)/4
	Next
End Function

Function ODE_CountJoints(body)
	Local obj.ODE_Object
	For obj=Each ODE_Object
		If obj\ODEbody=body Then Return BankSize(obj\geomBank)/4
	Next
End Function

Function ODE_GetGeom(body, index)
	Local obj.ODE_Object,geom.ODE_Geometry
	For obj=Each ODE_Object
		If obj\ODEbody=body
			geom=Object.ODE_Geometry(PeekInt(obj\geomBank,index*4))
			Return geom\ODEgeom
		EndIf
	Next
End Function

Function ODE_GetJoint(body, index)
	Local obj.ODE_Object
	For obj=Each ODE_Object
		If obj\ODEbody=body Then Return PeekInt(obj\jointBank,index*4)
	Next
End Function

Function ODE_AddSphere(body,radius#)
	Local newGeom
	
	newGeom=ODE_AddSphere_(body,radius)
	ODE_AddGeom_ body,newGeom
	
	Return newGeom
End Function

Function ODE_AddBox(body,sx#,sy#,sz#)
	Local newGeom
	
	newGeom=ODE_AddBox_(body,sx,sy,sz)
	ODE_AddGeom_ body,newGeom
	
	Return newGeom
End Function

Function ODE_AddCylinder(body,radius#,length#)
	Local newGeom
	
	newGeom=ODE_AddCylinder_(body,radius,length)
	ODE_AddGeom_ body,newGeom
	
	Return newGeom
End Function

Function ODE_AddCapsule(body,radius#,length#)
	Local newGeom
	
	newGeom=ODE_AddCapsule_(body,radius,length)
	ODE_AddGeom_ body,newGeom
	
	Return newGeom
End Function

Function ODE_AddTrimesh(triMesh)
	Local obj.ODE_Object,newGeom,tmData.ODE_Geometry
	Local surf,v,t,v0,v1,v2,ax#,ay#,az#,bx#,by#,bz#
	
	surf=GetSurface(triMesh,1)
	tmData=New ODE_Geometry
	
	tmData\vertexBank=CreateBank(CountVertices(surf)*12)
	For v=0 To CountVertices(surf)-1
		PokeFloat tmData\vertexBank,v*12,VertexX(surf,v)
		PokeFloat tmData\vertexBank,v*12+4,VertexY(surf,v)
		PokeFloat tmData\vertexBank,v*12+8,VertexZ(surf,v)
	Next
	
	tmData\triangleBank=CreateBank(CountTriangles(surf)*12)
	tmData\normalBank=CreateBank(CountTriangles(surf)*12)
	For t=0 To CountTriangles(surf)-1
		;Triangle normal code adapted from http://www.blitzbasic.com/codearcs/codearcs.php?code=466
		v0=TriangleVertex(surf,t,0)
		v1=TriangleVertex(surf,t,1)
		v2=TriangleVertex(surf,t,2)
		
		PokeInt tmData\triangleBank,t*12,v0
		PokeInt tmData\triangleBank,t*12+4,v1
		PokeInt tmData\triangleBank,t*12+8,v2
		
		ax=VertexX#(surf,v1)-VertexX#(surf,v0)
		ay=VertexY#(surf,v1)-VertexY#(surf,v0)
		az=VertexZ#(surf,v1)-VertexZ#(surf,v0)
		
		bx=VertexX#(surf,v2)-VertexX#(surf,v1)
		by=VertexY#(surf,v2)-VertexY#(surf,v1)
		bz=VertexZ#(surf,v2)-VertexZ#(surf,v1)
		
		PokeFloat tmData\normalBank,t*12,((ay#*bz#)-(az#*by#))
		PokeFloat tmData\normalBank,t*12+4,((az#*bx#)-(ax#*bz#))
		PokeFloat tmData\normalBank,t*12+8,((ax#*by#)-(ay#*bx#))
	Next
	
	newGeom=ODE_AddTrimesh_(0,tmData\vertexBank,CountVertices(surf),tmData\triangleBank,3*CountTriangles(surf),tmData\normalBank)
	;If body=0 Then 
	ODE_SetGeomMaterial_ newGeom,ODE_DefaultMaterial
	tmData\ODEgeom=newGeom
	
	Return newGeom
End Function

Function ODE_AddGeom_(body,newGeom)
	Local geom.ODE_Geometry=New ODE_Geometry,obj.ODE_Object
	geom\ODEgeom=newGeom
	If body=0
		ODE_SetGeomMaterial_ newGeom,ODE_DefaultMaterial
		geom\material=ODE_DefaultMaterial
	Else
		For obj=Each ODE_Object
			If obj\ODEbody=body
				ResizeBank obj\geomBank,BankSize(obj\geomBank)+4
				PokeInt obj\geomBank,BankSize(obj\geomBank)-4,Handle(geom)
				geom\material=obj\material
				geom\obj=obj
				Exit
			EndIf
		Next
	EndIf
End Function

Function ODE_FreeGeom(geom)
	Local body,i,gBankSize
	Local class,tmData.ODE_Geometry
	
	class=ODE_GetClass(geom)
	body=ODE_FreeGeom_(geom)
	
	For tmData=Each ODE_Geometry
		If geom=tmData\ODEgeom
			If class=ODE_TRIMESH_CLASS
				FreeBank tmData\vertexBank
				FreeBank tmData\triangleBank
				FreeBank tmData\normalBank
			EndIf
			Exit
		EndIf
	Next
	If tmData=Null Then RuntimeError "Geom does not exist!"
	
	If tmData\obj<>Null
		gBankSize=BankSize(tmData\obj\geomBank)
		For i=0 To gBankSize-4
			If PeekInt(tmData\obj\geomBank,i)=Handle(tmData)
				CopyBank tmData\obj\geomBank,i+4,tmData\obj\geomBank,i,gBankSize-(i+4)
				ResizeBank tmData\obj\geomBank,gBankSize-4
				Exit
			EndIf
		Next
	EndIf
	
	Delete tmData
	Return body
End Function

Function ODE_GeomPitch#(geom)
	ODE_GetRotation geom,2,private_ODE_MatrixBank
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
	Return EntityPitch(private_ODE_HelperPivot,1)
End Function

Function ODE_GeomYaw#(geom)
	ODE_GetRotation geom,2,private_ODE_MatrixBank
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
	Return EntityYaw(private_ODE_HelperPivot,1)
End Function

Function ODE_GeomRoll#(geom)
	ODE_GetRotation geom,2,private_ODE_MatrixBank
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,0),PeekFloat(private_ODE_MatrixBank,4),PeekFloat(private_ODE_MatrixBank,8),1
	AlignToVector private_ODE_HelperPivot,PeekFloat(private_ODE_MatrixBank,12),PeekFloat(private_ODE_MatrixBank,16),PeekFloat(private_ODE_MatrixBank,20),2
	Return EntityRoll(private_ODE_HelperPivot,1)
End Function

Function ODE_CreateJoint(jointType, bodyA, bodyB)
	Local newJoint, objA.ODE_Object, objB.ODE_Object
	
	For objA=Each ODE_Object
		If objA\ODEbody=bodyA Then Exit
	Next
	For objB=Each ODE_Object
		If objB\ODEbody=bodyB Then Exit
	Next
	If objA=Null Or objB=Null Then RuntimeError "Could not find rigid body!"
	
	newJoint=ODE_CreateJoint_(jointType, bodyA, bodyB)
	ResizeBank objA\jointBank,BankSize(objA\jointBank)+4
	PokeInt objA\jointBank,BankSize(objA\jointBank)-4,newJoint
	ResizeBank objB\jointBank,BankSize(objB\jointBank)+4
	PokeInt objB\jointBank,BankSize(objB\jointBank)-4,newJoint
	
	Return newJoint
End Function

Function ODE_FreeJoint(joint)
	Local bodyA,bodyB,objA.ODE_Object,objB.ODE_Object,jBankSize,i
	
	bodyA=ODE_GetJointBody(joint,0)
	If bodyA
		For objA=Each ODE_Object
			If objA\ODEbody=bodyA Then Exit
		Next
		If objA=Null Then RuntimeError "Could not find rigid body!"
		jBankSize=BankSize(objA\jointBank)
		For i=0 To jBankSize-4
			If PeekInt(objA\jointBank,i)=joint
				CopyBank objA\jointBank,i+4,objA\jointBank,i,jBankSize-(i+4)
				ResizeBank objA\jointBank,jBankSize-4
				Exit
			EndIf
		Next
	EndIf
	
	bodyB=ODE_GetJointBody(joint,1)
	If bodyB
		For objB=Each ODE_Object
			If objB\ODEbody=bodyB Then Exit
		Next
		If objB=Null Then RuntimeError "Could not find rigid body!"
		jBankSize=BankSize(objB\jointBank)
		For i=0 To jBankSize-4
			If PeekInt(objB\jointBank,i)=joint
				CopyBank objB\jointBank,i+4,objB\jointBank,i,jBankSize-(i+4)
				ResizeBank objB\jointBank,jBankSize-4
				Exit
			EndIf
		Next
	EndIf
	
	ODE_FreeJoint_ joint
End Function

Function ODE_RotateJoint(joint, pitch#, yaw#, roll#)
	RotateEntity private_ODE_HelperPivot,pitch,yaw,roll
	TFormPoint 0,0,1,private_ODE_HelperPivot,0
	ODE_SetJointAxis joint,TFormedX(),TFormedY(),TFormedZ()
End Function

Function ODE_GetConnector(bodyA, bodyB)
	Local objA.ODE_Object,objB.ODE_Object,i,j
	
	If ODE_Connected(bodyA, bodyB)=0 Then Return 0
	
	For objA=Each ODE_Object
		If objA\ODEbody=bodyA Then Exit
	Next
	If objA=Null Then RuntimeError "Could not find rigid body!"
	For objB=Each ODE_Object
		If objB\ODEbody=bodyB Then Exit
	Next
	If objB=Null Then RuntimeError "Could not find rigid body!"
	
	
	For i=0 To BankSize(objA\jointBank)-4
		For j=0 To BankSize(objB\jointBank)-4
			If PeekInt(objA\jointBank,i)=PeekInt(objB\jointBank,j) Then Return PeekInt(objA\jointBank,i)
		Next
	Next
	
	Return 0
End Function

;~IDEal Editor Parameters:
;~F#B#11#22#2A#32#39#62#94#A6#B2#C2#CD#E0#EB#104#10E#115#11C#123#12A
;~F#131#13B#142#14B#154#15D#166#195#1A8#1CA#1D1#1D8#1DF#1F3#219#21F
;~C#Blitz3D